#! /usr/bin/env python
#===============================================================================
# controllers.fitcontroller
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================
"""
.. _fitcontroller:

This module defines objets and methods related to the modelling 
and fitting procedure of the datasets

 """
from traits.api import (HasTraits, Str, Instance, on_trait_change,
                        Enum, Float, Int, Callable, Property,
                        cached_property, Button, Bool, Array, List,
                        DelegatesTo)
from traitsui.api import (View, Item, Label, TextEditor,
                          Group, HGroup, VGroup, spring,)
import os.path
import sys
import numpy as np
import re
from threading import Thread

from masai.modelling.parameters import ParameterScript
from masai.modelling.models import getmodel
from masai.modelling.sequence import PulseSequence
from masai.api import findregion
from masai.modelling import optimize # a modified copy of scipy.optimize
         
import logging
logger = logging.getLogger()

doc_params = """
#-----------------------------------------------------------
# syntax for parameters definition:
# name: value, low_bound,  high_bound
# *for fixed parameters
#  $ for variable parameters
#  > for reference to a parameter in the COMMON block
#    (> is forbidden in the COMMON block)
# common block parameters should not have a _ in their names
#-----------------------------------------------------------
"""

default_params_script = """
COMMON:
$ globgb: %(gb).3f, 0.0, none
$ globlb: 0.001, 0.0, none
$ globamp: %(ampl).3f, 0.0, none

MODEL: Line1
shape: voigtmodel
* amp:  1.0, 0.01, none
$ pos:  %(pos).3f, %(poslb).3f, %(poshb).3f
> gb:  globgb
> lb:  globlb

"""

doc_model = """
#-----------------------------
# TODO: make the documentation
#-----------------------------
"""

head_model = """
def userfunction(modeldata, pars):
 
#*******************************
"""

default_userscript = ""

tail_model = \
"""
#*******************************
    return modeldata
"""

#===============================================================================
# global function for modelfunctions
#===============================================================================

def parsing(expr, param):
    keyword = r"\b([a-z]+)\b" #match a whole word
    pat = re.compile(keyword)
    mo = pat.findall(expr)
    for kw in mo:
        if param.has_key(kw):
            expr = expr.replace(kw, str(param[kw]))
        elif np.__dict__.has_key(kw):# check if it is a recognized math function
            expr = expr.replace(kw, "np.%s" % kw)
    return expr
    
def prepare(param):
    """
    This function is needed for the script related to modelfunction
    """
    new_param = param.copy()
    
    for key in param.iterkeys():
        if param.reference[key]:
            new_param.reference[key] = False    # important to put it here 
                                                # before other instruction
            # try to interpret the given refpar expression
            refpar = param[key]
            while True:
                par = parsing(refpar, new_param)
                if par == refpar: break
                refpar = par

            try:                                    
                new_param[key] = eval(str(refpar))
            except:
                raise ValueError('Cannot evaluate the expression: %s: %s'
                                 % (key, param[refpar]))
                return
            new_param.fixed[key] = True
            new_param.reference[key] = True     #restore it for the next call 

    return new_param

#===============================================================================
# makefunc
#===============================================================================
def _makefunc(script=default_userscript):
    
    codescript = head_model
    scriptlines = script.split('\n')
    for line in scriptlines:
        codescript += "\t%s\n" % line
    codescript += tail_model
    
    #logger.debug(codescript)
    
    userfunction = None
    #try:
    code = compile(codescript, '<script>', 'exec')
    exec(code)
    return userfunction
    
    #except:
    #    return None

#===============================================================================
# automatic calculation of amplitude and baseline
#===============================================================================
def ampandbas(xi, expe, calc):
    """Automatically calculate correct amplitude A and baseline 
    (baseline linear model a*i+b) by detemining the zero of the first derivative
     with respect to A, a, and b
    """
    n = xi.size
    sic = sum(xi*calc)
    sc = sum(calc)
    sie = sum(xi*expe)
    se = sum(expe)
    sce = sum(expe*calc)
    sc2 = sum(calc*calc)
    A = (
         (12*n*sic + (-6*n**2 + 18*n - 12)*sc)*sie
         + (-6*n**2 + 18*n - 12)*se*sic 
         + (4*n**3 - 6*n**2 + 2*n)*sc*se
         + (-n**4 - 12*n**3 + 37*n**2 - 36*n + 12)*sce)/(12*n*sic**2
                                                + (-12*n**2 + 36*n - 24)*sc*sic
         + (-n**4 - 12*n**3 + 37*n**2 - 36*n + 12)*sc2 
         + (4*n**3 - 6*n**2 + 2*n)*sc**2)
    a = -(
          (12*n*sc2 - 12*sc**2)*sie + (12*sc*se - 12*n*sce)*sic 
          + (-6*n**2 + 18*n - 12)*sc2*se
          + (6*n**2 - 18*n + 12)*sc*sce)/(12*n*sic**2 
                                                + (-12*n**2 + 36*n - 24)*sc*sic
          + (-n**4 - 12*n**3 + 37*n**2 - 36*n + 12)*sc2 
          + (4*n**3 - 6*n**2 + 2*n)*sc**2)
    b = -(
          (12*sc*sic + (-6*n**2 + 18*n - 12)*sc2)*sie
          - 12*se*sic**2 + (6*n**2 - 18*n + 12)*sce*sic 
          + (4*n**3 - 6*n**2 + 2*n)*sc2*se
          + (-4*n**3 + 6*n**2 - 2*n)*sc*sce)/(12*n*sic**2
          + (-12*n**2 + 36*n - 24)*sc*sic 
          + (-n**4 - 12*n**3 + 37*n**2 - 36*n + 12)*sc2
          + (4*n**3 - 6*n**2 + 2*n)*sc**2)
    # in case the modeldata is zero, to avoid further errors
    if np.isnan(A): A=0.0
    if np.isnan(a): a=0.0
    if np.isnan(b): b=0.0
    return A, a, b    

#===============================================================================
# ModelCompute
#===============================================================================
class ModelComputeThread(Thread):
    """Inspired from gael varoquaux tutorial
    
    """
    def run(self):
        control = self.control
        if not control.sequence:
            return  # not fully initialized yet, return
        try:
            logger.debug(str(self)+' started')
            control.lastfitfailure = None
            modeldata, names = control._get_modeldata() 
        except Exception, e:
            control.lastfitfailure = e
            control.paramscript.lastparfailure = e
            logger.error(e)
            logger.debug(str(self)+' stopped on error')
            return
            
        self.control.fd.modelnames = names   # must come first 
                                            #(to avoid plotting routine errors)
        self.control.fd.modeldata = modeldata
        logger.debug(str(self)+' finished')
        
    def stop(self):
        logger.debug(str(self)+' stopped')
        self._Thread__stop()

#===============================================================================
# ModelFitting
#===============================================================================
class ModelFittingThread(Thread):
    """Inspired from gael varoquaux tutorial
    """
    stopped = False
    def run(self):
        control = self.control
        try:
            control.lastfitfailure = None
            control._fitting()
        except Exception, e:
            #some common errors
            if e.message.strip() == \
                       "zero-size array to maximum.reduce without identity":
                control.lastfitfailure = ValueError(
                           "At least one parameter should be fitted\n"\
                           + "(use $ in front of the parameter to fit)")
            else:
                control.lastfitfailure = e
            logger.error(e)
            
        control._running = False
             
#===============================================================================
# Class FitController
#===============================================================================
class FitController(HasTraits):
    
    fd = Instance('masai.fileio.source.Source')
    
    paramscript = Instance(ParameterScript) 
    userscriptfile = Str('')
    userscriptcontent = Str
    userfunction = Property(Callable,
                            depends_on='userscriptcontent, fd.is_resized')
    lastfitfailure = Instance(Exception, allow_none=True)
    computethread = Instance(ModelComputeThread)
    fittingthread = Instance(ModelFittingThread)
    
    sequence = Instance(PulseSequence)
    
    startbtn = Button('START FIT')
    stopbtn = Button('STOP FIT')

    fp = Property(depends_on='paramscript.updated')
    fittable = DelegatesTo('fd')
    
    method = Enum(("SIMPLEX"))#, "BFGS", "CG"))
    tolerance = Float(1.e-5)
    maxiter = Int(250)
    maxfuncalls = Int(2500)
    plotevery = Int(10)
    
    funcalls = Int
    iterations = Int
    fopt = Float
    warnmess = Str
    fitpar = Str
    
    intensities = Str('')
    A = DelegatesTo('fd')
        
    #===========================================================================
    # private
    #===========================================================================        
    _loading = Bool(True)
    _running = Bool(False)
    
    _save_modeldata = Array(allows_none=True)
    _save_names= List
    _save_fp = Instance('masai.modelling.parameters.FitParameters')     
      
    #===========================================================================
    # default_traits_view
    #===========================================================================
    def default_traits_view(self):        
        return View(
                VGroup(
                    Group(
                        Group(Item('paramscript',
                                    show_label=False,
                                    style='custom',
                                    #enabled_when='not _running'
                                ),
                               label="Parameters",
                               dock='tab',
                        ),
                        Group(
                              Item(name='intensities',
                                   show_label = False,
                                   style = 'readonly',
                                   editor=TextEditor(multi_line=True)
                                   ),
                              label= 'Report',
                              dock = 'tab',
                              ),
                        #=======================================================
                        ## NOT USED FOR NOW
                        # Group( Item('userscriptcontent',
                        #             show_label=False,
                        #             style='custom',
                        #             editor=CodeEditor(lexer='python', 
                        #                               auto_set=True)),
                        #       label="User script",
                        #       dock='tab',
                        # ),
                        # Group( Item('sequence',
                        #             style='custom',
                        #             show_label =False),
                        #       label="pulse sequence",
                        #       dock='tab',
                        # ),
                        #=======================================================
                        springy=True,
                        layout='tabbed',
                    ),
                    Group(
                        VGroup(
                               #Item('method', label='Method',), #not used for now
                               Item('tolerance', width=50),
                               Item('maxiter',
                                    label="Max of iterations",
                                    width=50),
                               Item('maxfuncalls',
                                    label='Max of calls/iterations',
                                    width=50),
#                               Item('plotevery',
#                                    label='Nb of iterations/plot updates',
#                                    width=50),
                               orientation='vertical',
                        ),
                        VGroup(
                               HGroup(
                               Item('iterations',
                                    label=u"iterations",
                                    style='readonly',
                                    width=50),
                               Item('funcalls',
                                    label='calls',
                                    style='readonly',
                                    width=50),
                               Item('fopt',
                                    label='Error (Chi)',
                                    style='readonly',
                                    width=100),
                               ),
                               Item('warnmess',
                                    editor=TextEditor(multi_line=True),
                                    show_label=False,
                                    width=300,
                                    style='readonly',
                                    emphasized=True,
                                    visible_when="funcalls>0",),
                               Item('lastfitfailure',
                                    editor=TextEditor(multi_line=True),
                                    show_label=False,
                                    visible_when='lastfitfailure',
                                    style='readonly',
                                    width=350,
                                    emphasized=True),
                               orientation='vertical',
                               show_border=True,
                               label='results',
                        ),
                        Item('_'),
                        HGroup(
                              Item('startbtn', show_label=False,
                                    visible_when='not _running',
                                    ),
                               Item('stopbtn', show_label=False,
                                    visible_when='_running'),
                               spring,
                               orientation='horizontal',
                        ),
                        orientation='vertical',
                    ),
                    visible_when='fittable',
                ),
                 VGroup(
                   Item('_'),
                   Label('WARNING'),
                   Label('A fourier transform must be done to access fitting'),
                   Item('_'),
                   visible_when='not fittable or fd is None',
                ),
                resizable=True,
                scrollable=True,
            )
        
    #===========================================================================
    # fd event router
    #===========================================================================
    @on_trait_change('fd.+')
    def _eventrouter(self, objet=None, name=None, old=None, new=None):
        
        if not self.fd.is_loaded: 
            # we return if the source is not valid
            return

        if name == 'fd':
            logger.debug("initialisation of Fit class")

#            self.model = ''
#            self.simpson = ''

            self.paramscript = ParameterScript(fd=self.fd)

#            # set userscriptcontent       
            if not os.path.exists(self.userscriptfile):
                self.userscriptcontent = default_userscript
            else:
                self.userscriptfile = os.path.join(self.fd._directory, 
                                               "userscript.fit")
                with open(self.userscriptfile, 'rb') as f:
                    self.userscriptcontent = f.read()
        
            _type = self.fd.sourcepref.get("fit.sequence", 'ideal_pulse')  
            self.sequence = PulseSequence(type=_type)


    @cached_property
    def _get_fp(self):
        return self.paramscript.fp
           
    def _save_userscript(self):
        if self.userscript:
            f = open(self.userscriptfile, 'wb')
            f.write(self.userscriptcontent)
            f.close()           
    
    @on_trait_change('userscriptcontent')
    def _update_userscriptcontent(self, name, old, new):
        self._save_userscript()
        self.updated = True
          
    @cached_property
    def _get_userfunction(self): 
        """user function 
        """
        #logger.debug('get userfunction')
        #create or update the user function
        userfunction = _makefunc(self.userscriptcontent)
        return userfunction

    #===========================================================================
    # _get_modeldata
    #===========================================================================
    def _get_modeldata(self):
        """Calculate the model
        """
        
        if not self.fd.forceupdatemodel and \
                         hasattr(self, '_save_fp') and self.fp == self._save_fp:
            # absolutely no changes, just return the saved modeldata
            return self._save_modeldata, self._save_names
        
        self.fd.forceupdatemodel=False # suppress force update
        self._save_fp = self.fp.copy()

        # Prepare parameters
        parameters = prepare(self.fp)

        # Get the list of models
        models = self.fp.models
        nbmodels = len(models)

        # Make an array 'modeldata' with the size of the source of data 
        # which will contains the data produced by the models
        # This name must always be 'modeldata' 
        # which will be returned to the main program.
        
        uc = self.fd.units
        x = uc[-1].ppm_scale()
        modeldata = np.zeros((nbmodels + 2, x.size)).astype(float)
    
        if nbmodels < 1:
            names = ['baseline', 'modelsum']
            self._save_modeldata = modeldata
            self._save_names = names
            return modeldata, names

        # Calculates model data
        # The first row (i=0) of the modeldata array is the baseline, 
        # so we fill the array starting at row 1
        row = 0
        names = ['baseline', ]
        for model in models:
            if not model.startswith('baseline'):
                row += 1
                modeldata[row] = getmodel(x, model, parameters,
                                          fd=self.fd, sequence=self.sequence)
                names.append(model)
            else:
                modeldata[0] += getmodel(x, model, parameters,
                                         fd=self.fd, sequence=self.sequence)
                
        # call the userfunction 
        # (in case something is executed in it - nothing by default)
        if default_userscript:
            modeldata = self.userfunction(modeldata, self.fp)  
            
        # make the sum
        modeldata[row + 1] = modeldata.sum(axis=0)
        names.append('modelsum')

        #remove unused row
        modeldata = modeldata[:row + 2]
        
        #calculate automatic baseline and amplitude
        data = self.fd.data.real
        xi = np.arange(x.size)
        
        # take into account fitzone
        uc = self.fd.units
        x = uc[-1].ppm_scale()
        fitzone = findregion(x, self.fd.zone)

        A, a, b = ampandbas(fitzone-fitzone[0], data.take(fitzone), 
                            modeldata[-1].take(fitzone))
        
        modeldata = A*modeldata
        baseline = a*(xi-fitzone[0]) + b
        
        #update the modeldata
        modeldata[0] += baseline
        modeldata[-1] += baseline
        
        self._save_modeldata = modeldata
        self._save_names = names
        self.A = A
        
        # calculate intensities
        at = 0.
        for model in models:
            if model.startswith('baseline'):
                continue
            at += parameters["ampl_%s"%model]
            
        self.intensities = 'Intensities:\n       \tppm   \tRelative (%)     \tAbsolute\n'
        for model in models:
            if model.startswith('baseline'):
                continue
            self.intensities += \
                '%(model)7s:\t%(pos).2f\t%(ampl)g \t%(conc)g\n'%{
                                  'model':model,
                                  'pos':parameters["pos_%s"%model],
                                  'ampl':parameters["ampl_%s"%model]*100./at,
                                  'conc':parameters["ampl_%s"%model]*self.A}
                

        return modeldata, names
    
    #===========================================================================
    # recompute models
    #===========================================================================
    @on_trait_change('fp, fd.is_resized, fd.forceupdatemodel')
    def _updated_fp(self, objet=None, name=None, old=None, new=None):
                
        self.lastfitfailure = None
        if hasattr(self.paramscript,'lastparfailure') \
                   and self.paramscript.lastparfailure: 
            return
        
        #check is the thread is already running. If yes send abort first
        if self.computethread and self.computethread.isAlive():
            self.computethread.stop()
        self.computethread = ModelComputeThread()
        self.computethread.control = self
        self.computethread.start()
        
        self.fd.model_need_replot

    #===========================================================================
    # Fitting actions
    #===========================================================================
    def _startbtn_fired(self):
        """
        Start fitting 
        """
        logger.info('\nfit started...')
        logger.info(
                'Take a cup of coffee and be patient as it may take a while!')
        
        self.fd.can_plot=False
        self.funcalls = 0
        self._running = True
        self.warnmess = ''
        #self.lastfitfailure = ''
        
        if len(self.fd.zone) < 1:
            self._running = False
            return  # this should not happen
 
        
        #check is the thread is already running. If yes send abort first
        if self.fittingthread and self.fittingthread.isAlive():
            self.fittingthread.wants_abort = True
        else:
            self.fittingthread = ModelFittingThread()
            self.fittingthread.wants_abort = False
            self.fittingthread.control = self
            self.fittingthread.start()       

    def _stopbtn_fired(self):
        """
        Stop fit  : TODO make it working using Threads
        """
        if self.fittingthread and self.fittingthread.isAlive():
            self.fittingthread.stopped = True
        self.fd.can_plot=True
        #logger.info('fit stopped. Wait for the last execution call')
        return True
                
    #===========================================================================
    # _fitting
    #===========================================================================

    #  Internal/external transformation
    #  These transfomations are used in the MINUIT package, 
    #  and described in detail 
    #  in the section 1.3.1 of the MINUIT User's Guide.
    
    def internal(self, pe, lob, upb):
        islob = (lob > -0.1/sys.float_info.epsilon) 
        isupb = (upb < +0.1/sys.float_info.epsilon)
        
        if islob and isupb:
            # With min and max bounds defined
            pi = np.arcsin((2*(pe - lob)/(upb - lob)) - 1.)
        elif isupb:
            # With only max defined
            pi = np.sqrt((upb - pe + 1.)**2 - 1.)
        elif islob:
            # With only min defined
            pi = np.sqrt((pe - lob + 1.)**2 - 1.)
        else:
            pi = pe    
        return pi
    
    def external(self, pi, lob, upb):
        islob = (lob > -0.1/sys.float_info.epsilon) 
        isupb = (upb < +0.1/sys.float_info.epsilon)

        if islob and isupb:
            #  With min and max bounds defined
            pe = lob + ((upb - lob)/2.)*(np.sin(pi) + 1.)
        elif isupb:
            #With only max defined
            pe = upb + 1. - np.sqrt(pi**2 + 1.)
        elif islob:
            #With only min defined
            pe = lob - 1. + np.sqrt(pi**2 + 1.)
        else:
            pe = pi
        return pe
 
    def _fitting(self):   
        """Main fitting procedure
        """ 
        logger.info('Entering fitting procedure') 
        self._forcereplot = False
        
        uc = self.fd.units
        x = uc[-1].ppm_scale()
        fitzone = findregion(x, self.fd.zone)
                 
        # internally defined function chi2
        def chi2(params, keys):
            """ 
            Return sum((y - x)**2) 
            
            """
            for i in range(len(params)):
                key = keys[i]
                # restore external parameters in case of bounding
                pe = self.external(params[i],
                                   self.fp.lob[key], self.fp.upb[key])
                self.fp[key] = pe, self.fp.lob[key], self.fp.upb[key]

            # model spectrum
            modeldata = self._get_modeldata()[0]
                
            # baseline is already summed with modeldata[-1]
            diff = (self.fd.data.real - modeldata[-1])**2 
            chi2 = np.sum(diff.take(fitzone))

            return chi2

        p = []
        keys = []

        for key in self.fp.keys():
            if not self.fp.fixed[key]:
                # we make internal parameters in case of bounding
                p.append(self.internal(self.fp[key],
                                       self.fp.lob[key], self.fp.upb[key]))
                keys.append(key)    

        res, self.fopt, self.funcalls, self.warnmess = self.minimize(chi2,
                                                        p, args=(keys,))

        self.fopt = np.sqrt(self.fopt)
        #res is internal parameters
        #restore the external parameter
        for i in range(len(p)):
            key = keys[i]
            pe = self.external(res[i], self.fp.lob[key], self.fp.upb[key])
            self.fp[key] =( pe, self.fp.lob[key],
                                self.fp.upb[key], self.fp.fixed[key])

        # update the paramscript content
        self.paramscript.content = str(self.fp)
        self._running = False
        self.fd.can_plot=True
        return

    def _callforresults(self, params, keys, funcall, iteration, fopt):
        """call back function called by minimize
        """
        for i in range(len(params)):
            key = keys[i]
            # restore external parameters in case of bounding
            pe = self.external(params[i], self.fp.lob[key], self.fp.upb[key])
            self.fp[key] = pe, self.fp.lob[key], self.fp.upb[key]

        # update parameter script
        self.funcalls = funcall
        self.iterations = iteration
        self.fopt = np.sqrt(fopt)

        return self.fittingthread.stopped   

    #===========================================================================
    # minimize
    #===========================================================================
    def minimize(self, func, x0, args=()):
        """
        """
        
        _epsilon = np.sqrt(np.finfo(float).eps)
        xtol = ftol = self.tolerance
        maxiter = self.maxiter
        maxfun = self.maxfuncalls
        callback = self._callforresults
        
        #allvecs = None
        
        if self.method == 'SIMPLEX':
            res = optimize.fmin(func, x0, args=args, xtol=xtol,
                                ftol=ftol, maxiter=maxiter,
                                maxfun=maxfun, full_output=True, disp=True,
                                retall=True, callback=callback)
            return res[0], res[1], res[3], res[4]

#        elif self.method == 'BFGS':
#            res = optimize.fmin_bfgs(func, x0, args=args, gtol=ftol, 
# norm=np.Inf,
#                      epsilon=_epsilon, maxiter=maxiter, 
# full_output=True, disp=True,
#                       retall=True, callback=callback)
#            #res = (xopt, {fopt, gopt, Hopt, func_calls, grad_calls, 
# warnflag}, <allvecs>)
#            return res[0]
#        
#        elif self.method == 'CG':
#            res = optimize.fmin_cg(func, x0, args=args, gtol=ftol, norm=np.Inf,
#                      epsilon=_epsilon, maxiter=maxiter, full_output=1, disp=1,
#                      retall=1, callback=callback)
#            return res[0]
        
        else:
            raise NotImplementedError, "Method not implemented"     
        



#! /usr/bin/env python
#===============================================================================
# controllers.plotcontroller
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================
"""
.. _plotcontroller:

This module defines objets and methods related to the plotting of the
datasets.

"""
import wx

#===============================================================================
# Enthought imports
#===============================================================================
from traits.api import (HasTraits, Instance, Range, Bool, Int, Enum,
                        Float, Button, on_trait_change, Color, ListFloat,
                        List, Property, Str,
                        DelegatesTo)
from traitsui.api import (View, Item, Group, VGroup, HGroup, HFlow,
                          spring, ScrubberEditor, ColorEditor,
                          Handler, EnumEditor)

#===============================================================================
# numpy, mathplotlib
#===============================================================================
import numpy as np
from masai.views.viewplot import mplcolor, lstyle

#===============================================================================
# logger
#===============================================================================
import logging
logger = logging.getLogger()

#===============================================================================
# masai imports
#===============================================================================

#===============================================================================
# PlotController
#===============================================================================
class PlotController(HasTraits):
    """Control viewplot appearance and behaviour 

    """
    # the source of data
    fd = Instance('masai.api.Source')
    viewplot = Instance('masai.views.ViewPlot')

    show_model = DelegatesTo('fd')
    show_diff = DelegatesTo('fd')
    autoscale = DelegatesTo('fd')
    subtract = DelegatesTo('fd')
    fittable = DelegatesTo('fd')
    unmodified = DelegatesTo('fd')
    
    data_width = Range(.1, None, .8)
    data_color = Color(wx.Colour(128, 128, 128, 255))
    data_style = Enum(lstyle.keys())
    data_offset = Float(.0)

    diff_width = Range(.1, None, .8)
    diff_color = Color(wx.Colour(128, 128, 128, 255))
    diff_style = Enum(lstyle.keys())
    diff_offset = Float(.0)

    modelnames = DelegatesTo('fd')
    line = Str('baseline')

    width = Range(.1, None, .8)
    color = Color
    style = Enum('Solid', lstyle.keys())
    offset = Float(.0)

    grpexp = VGroup(
                    HGroup(
                           Item('data_style', label='style', width=30),
                           Item('data_color', label='color', width=100,
                                editor=ColorEditor()),
                           ),
                    HGroup(
                           Item('data_width', label='width', width=30,
                                editor=ScrubberEditor(increment=.1,
                                                      hover_color=0xFFFFFF,
                                                      active_color=0xB0DD8E),
                                ),
                           Item('data_offset', label='offset(%)', width=30,
                                editor=ScrubberEditor(increment=.5,
                                                      hover_color=0xFFFFFF,
                                                      active_color=0xB0DD8E),
                                ),
                           ),
                    Item('subtract', label='Subtract baseline',
                         visible_when='modelnames and fittable'),
                    Item('unmodified', label='show original data',
                         visible_when = 'not fittable'),
                    label='Experimental',
                    show_border=True,
                    )

    grpmod = VGroup(
                    Item(name='line',
                         editor=EnumEditor(name='modelnames'),
                         ),
                    VGroup(
                           HGroup(
                                  Item(name='style',width=30),
                                  Item(name='color', width=100,
                                       editor=ColorEditor()
                                       ),
                                  ),
                           HGroup(
                                  Item(name='width', width=30,
                                       editor=ScrubberEditor(increment=.1,
                                                        hover_color=0xFFFFFF,
                                                        active_color=0xB0DD8E),
                                       ),
                                  Item(name='offset',
                                       label='offset(%)', width=30,
                                       editor=ScrubberEditor(increment=.5,
                                                        hover_color=0xFFFFFF,
                                                        active_color=0xB0DD8E),
                                       ),
                                  ),
                           enabled_when='not (subtract and line=="baseline")',
                           ),
                    label='Models',
                    show_border=True,
                    visible_when='fittable',
                    ),

    grpdif = VGroup(
                    HGroup(
                           Item('diff_style', label='style', width=30),
                           Item('diff_color', label='color', width=100,
                                editor=ColorEditor()
                                ),
                           ),
                    HGroup(
                           Item('diff_width', label='width', width=30,
                                editor=ScrubberEditor(increment=.1,
                                                      hover_color=0xFFFFFF,
                                                      active_color=0xB0DD8E),
                                ),
                           Item('diff_offset',
                                label='offset(%)', width=30,
                                editor=ScrubberEditor(increment=.5,
                                                      hover_color=0xFFFFFF,
                                                      active_color=0xB0DD8E),
                                ),
                           ),
                    label='Difference',
                    show_border=True,
                    visible_when="show_diff and modelnames and fittable"
                    )

    grpsho = HGroup(
                    Item("show_model", label="Show model"),
                    Item("show_diff", label="Show difference",
                         visible_when='show_model',
                         ),
                    visible_when='fittable',
                    label='Model',
                    show_border=True,
                    )

    grpmis = HGroup(
                    Item("autoscale", label='Auto scale y axis?',
                         visible_when='fd.ndims==1'),
                    label='Miscellaneous',
                    show_border=True,
                    )
    
    view = View(
                Group(
                      grpexp,
                      grpmod,
                      grpdif,
                      grpsho,
                      grpmis,
                      orientation='vertical',
                      ),
                width=350,
                resizable=True,
                )

    def _line_default(self):
        return 'baseline'

    def _color__default(self):
        return self.fd.line_attributes[self.line][2]

    def _style_default(self):
        return self.fd.line_attributes[self.line][0]

    def _width_default(self):
        return self.fd.line_attributes[self.line][1]

    def _offset_default(self):
        return self.fd.line_attributes[self.line][3]  

    def _anytrait_changed(self, name, old, new):
        """Detect all changes in the plotcontroller traits
        """
        logger.debug ("%s, %s, %s" % (name, str(old), str(new)))

        if self.viewplot is None or self.fd is None:
            return
        
        if name == 'data_style':
            self.fd.line_attributes['experimental'][0] = self.data_style
            self.viewplot.updatelineattr('experimental')

        if name == 'data_width':
            self.fd.line_attributes['experimental'][1] = self.data_width
            self.viewplot.updatelineattr('experimental')

        if name == 'data_color_':
            self.fd.line_attributes['experimental'][2] = self.data_color_
            self.viewplot.updatelineattr('experimental')

        if name == 'data_offset':
            self.fd.line_attributes['experimental'][3] = self.data_offset
            self.viewplot.updatelineattr('experimental')

        if name == 'diff_style':
            self.fd.line_attributes['difference'][0] = self.diff_style
            self.viewplot.updatelineattr('difference')

        if name == 'diff_width':
            self.fd.line_attributes['difference'][1] = self.diff_width
            self.viewplot.updatelineattr('difference')

        if name == 'diff_color_':
            self.fd.line_attributes['difference'][2] = self.diff_color_
            self.viewplot.updatelineattr('difference')

        if name == 'diff_offset':
            self.fd.line_attributes['difference'][3] = self.diff_offset
            self.viewplot.updatelineattr('difference')

        if name == 'fd':
            # initialization 
            self.data_style = self.fd.line_attributes['experimental'][0]
            self.data_width = self.fd.line_attributes['experimental'][1]
            # warning data_color must be followed by a _
            # if not the property is not set (good to know)
            self.data_color_ = self.fd.line_attributes['experimental'][2]
            self.data_offset = self.fd.line_attributes['experimental'][3]
            self.viewplot.updatelineattr('experimental')

            self.diff_style = self.fd.line_attributes['difference'][0]
            self.diff_width = self.fd.line_attributes['difference'][1]
            self.diff_color_ = self.fd.line_attributes['difference'][2]
            self.diff_offset = self.fd.line_attributes['difference'][3]
            self.viewplot.updatelineattr('difference')

        if name == 'line':
            line = self.line
            self.color_ = self.fd.line_attributes[line][2]
            self.color = self.fd.line_attributes[line][2]
            self.offset = self.fd.line_attributes[line][3]
            self.style = self.fd.line_attributes[line][0]
            self.width = self.fd.line_attributes[line][1]
            self.viewplot.updatelineattr(self.line)     

        if name == 'style':
            self.fd.line_attributes[self.line][0] = self.style
            self.viewplot.updatelineattr(self.line)

        if name == 'width':
            self.fd.line_attributes[self.line][1] = self.width
            self.viewplot.updatelineattr(self.line)

        if name == 'color_':
            self.fd.line_attributes[self.line][2] = self.color_
            self.viewplot.updatelineattr(self.line)

        if name == 'offset':    
            self.fd.line_attributes[self.line][3] = self.offset
            self.viewplot.updatelineattr(self.line)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#===============================================================================
# processing.process
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================
"""
.. _process:

This module contains the processing commands.

"""
import sys

#===============================================================================
# logger
#===============================================================================
import logging
logger = logging.getLogger()

import numpy as np

from scipy.optimize import optimize

from processbase import _em, _pk, _ft, _ift, _zf
 
# definitions

# MC2 or FnMODE table
undefined = -1
QF = 0
QSEQ = 1
TPPI = 2 
STATES = 3
STATESTPPI = 4
ECHOANTIECHO = 5

# save stdout
STDOUT = sys.stdout
# save stderr
STDERR = sys.stderr

__all__ = ['bc', 'tdeff','em', 'azf','zf', 'reducesampling', 'clipping', 'cursor2max', 'pk', 'apk',
           'ft', 'ift', 'efp', 'xf2', 'xf1', 'findregion', 'largest_power_of_2', 'transpose', 'makestates', 
           'shearing_ratio']
#===============================================================================
# Blackhole
#===============================================================================
class Blackhole: 
    """Utility class used to suppress some output
    """
    def write(self, *msg): 
        pass
    
#===============================================================================
# apodization, sampling and zerofilling
#===============================================================================
def bc(fd):
    """Correct fid offsets
    
    if BC_mod parameter is set, a baseline correction is applied to the fid
    in the direct (or indirect if transposed data) dimension. 
    The average value of the last quarter of the fid 
    is substracted to the whole data

    data ndarray are modified in place

    .. warning:: 
    
        fd.data should ideally be some time domain data, i.e., 
        is_freq set to false or wrong results can occurs
        
    .. todo:: 
    
        Only quad bc correction mode is setup: handle different mode of BC correction.   
    
    Parameters
    ----------
    fd : Source instance
        Contains bruker NMR parameters and data 
    

        
    Examples
    --------
        
        >>> from masai.api import *
        >>> fd = Source(path = 'test/1')
        >>> bc(fd)
        
        or for 2D
        
        >>> fd= Source(path='test/2')
        >>> bc(fd)

            
    """
    logger.debug("bc")
    size = fd.data.shape[-1]
    if fd.BC_mod > 0: 
        bc = np.mean(fd.data[..., -size / 4:], axis= -1)
        fd.data = np.transpose(np.transpose(fd.data) - bc)
     
    return

def tdeff(fd):
    """
    reduce the effective time domain size (td).
    
    """
    if fd.is_freq[-1]:
        return 

    logger.debug("tdeff")
    
    if fd.TDeffSAV[-1]>0.0 and (fd.TDeff[-1] > fd.TDeffSAV[-1] or fd.TDeff[-1] > fd.TD[-1]):
        logger.error("TDeff was not applied: it was already performed on a shorter size or it is geater than TD")
        return
    
    fd.data = fd.data[:fd.TDeff[-1]] 
    
    fd.TDeffSAV[-1] = fd.TDeff[-1]
    
    return

def em(fd, lb=None):
    """
    Exponential apodization
    
    The command em performs an exponential window multiplication of the fd. 
    It is the most used window function for NMR spectra. 
    em multiplies each data point :math:`k` with the factor:
    
    .. math::
        em(k) = \\exp\\left[ -\\pi  \\, \\frac{k \\, LB}{2 \\, SW\_h}\\right]
        
    where LB (the line broadening factor in Hz) is a processing parameter 
    and SW_h (the spectral width) an acquisition status parameter.
    
    
    Parameters
    ----------
    fd : Source instance
        Contains bruker NMR parameters and  data 
    lb : float, optional, None by default
        Line broadening in Hz, replace the value stored in fd if specified
    


    Examples
    --------
        
        >>> from masai.api import *
        >>> fd = Source(path = 'test/1')
        >>> em(fd)
        
        or for 2D
        
        >>> fd= Source(path='test/2')
        >>> em(fd)

    """
    if fd.is_freq[-1]:
        return
    
    sw = fd.SW_h[-1]
    if lb is not None: fd.LB[-1] = lb
    lb = fd.LB[-1]- fd.LBSAV[-1]  # Inverse exponential width
    
    if abs(lb)>0.0 : 
        logger.debug("em")
        fd.data = _em(fd.data, lb / sw / 2.) 
    
    fd.LBSAV[-1] = fd.LB[-1]
    
    return

def azf(fd):
    """ 
    Automatic zero filling to two times the next largest power of two
    
    Parameters
    ----------
    fd : Source instance
        Contains bruker NMR parameters and  data
    


    Examples
    --------
        
        >>> from masai.api import *
        >>> fd = Source(path = 'test/1')
        >>> azf(fd)
        
        or for 2D
        
        >>> fd= Source(path='test/2')
        >>> azf(fd)

    """
    logger.debug("azf")
    
    size = fd.data.shape[-1]
    size = largest_power_of_2(size)
    zf(fd, size * 2)
    
    return

def zf(fd, size):
    """ 
    zero filling to the specified size.
    
    Parameters
    ----------
    fd : Source instance
        Contains bruker NMR parameters and  data
    size : Int
        Desired new data size after zero filling
        

        
    Examples
    --------
        
        >>> from masai.api import *
        >>> fd = Source(path = 'test/1')
        >>> zf(fd, 2**14)
        
        or for 2D
        
        >>> fd= Source(path='test/2')
        >>> zf(fd, 2**14)


    """
    logger.debug("zf")
 
    fd.data = _zf(fd.data, size)
    
    return

def reducesampling(fd, sampling):
    """Divide the sampling rate by the sampling parameter
    
    Undersampling can be used to reduce the size of the spectral window, e.g,
    
    * if sampling=3, then data are taken every three points in the direct dimension, 
    * or if sampling=(3,2), then additionnaly, a double sampling is performed in the second dimension
    
    works always on the original data (suppressing previous apodization for instance). So it should be the first 
    action to perform on a fid.
    
    Parameters
    ----------
    fd : Source Instance
        Should be time domain data
    sampling : int or tuple of int 
        Divider to resample the fd.data ndarray
        The size of the Tuple should be fd.data.shape
        
    Examples
    --------
        
        >>> from masai.api import *
        >>> fd = Source(path = 'test/1')
        >>> reducesampling(fd, 6)
        
        or for 2D
        
        >>> fd= Source(path='test/2')
        >>> reducesampling(fd, 6)

    """
    if fd.is_freq[-1]:
        # this apply only to time domain data
        return
    logger.debug("reducesampling")
    
    if isinstance(sampling, tuple):
        sampling_2 = sampling[0]
        sampling_1 = sampling[1]
    else:
        sampling_2 = sampling
        sampling_1 = 1.0
        
    if sampling_2 > 0:
        fd.data = fd.data_orig[..., 0::1 * sampling_2]
        size = fd.data.shape[-1]
        fd.SW[-1] = fd.SW[-1] * size / fd.TD[-1]
        fd.SW_h[-1] = fd.SW_h[-1] * size / fd.TD[-1]
        fd.TD[-1] = size
        fd.is_resized = True
     
    #we need to set the position of the phcursor
    fd.phcursor = cursor2max(fd)
       
    if sampling_1 > 1:
        # warning here this strongly depends on the acquisition mode
        # fd.data = fd.data[0::1*sampling_2]
        logger.warn("Warning: reduce sampling in indirect dimension not done. Not yet implemented")
        return False

    return

def clipping(fd):
        """
        reduce the size of the transformed spectra to a smaller range (clipping region).
        The number of points is however made equal to a power of two.
        
        Parameters
        ----------
        fd : Source Instance
            Should be the frequency domain data
            clipping region is known from the clip attribute of fd
        
        """
        if not fd.is_freq[-1]:
            # this apply only to FT transformed data
            return
    
        if not fd.clip:
            return
        
        # we make a temporary copy of the source as we do not want refreshing during
        fd_ = fd.copy()
        
        logger.debug('clipping')
        uc = fd.units
        x = uc[-1].ppm_scale()
        xlim = uc[-1].ppm_limits()
        if list(xlim)==fd.clip:
            return
        clipped = findregion(x, fd.clip)
        
        if len(clipped)%2!=0:
            # make it even
            clipped.append(clipped[-1]+1)
        
        shape = list(fd_.data.shape)
        shape[-1] = len(clipped)
        newdata = fd_.data.take(clipped).flatten()
        fd_.data = newdata.reshape(shape)

        # correct scaling units
        xhigh, xlow = float(x[clipped[-1]]) , float(x[clipped[0]])
        fd.SW[-1] = xhigh - xlow 
        fd.SW_h[-1] = fd.SW[-1] * fd.SFO1[-1]
        sr = (xlow + fd.SW[-1] / 2.) * fd.SFO1[-1]
        fd.SF[-1] = fd.SFO1[-1] - sr * 1.e-6
                
        # make the spectrum number of point a power of two (necesary for further simulation using simpson for instance)
        fd_.data.imag = 0.0       # we use a hilbert transform to avoid problem with edge artifacts
        fd_.is_freq[-1]=False
        fd_.data = _ift(fd_.data)
        fd_.data = fd_.data[:len(fd_.data)/2]
        azf(fd_)
        fd_.data = _ft(fd_.data)
        # correct for some baseline while returning the data
        if fd_.ndims>1:
            #TODO : check this
            dataT = fd_.data.T 
            fd_.data = (dataT - (np.mean(dataT[0:100])+np.mean(dataT[-100:-1]))/2.).T
        else:
            fd_.data = fd_.data - (np.mean(fd_.data[0:100])+np.mean(fd_.data[-100:-1]))/2.


        fd_.is_freq[-1] = True
        
        # restore to the original data
        fd.data = fd_.data.copy()
        
        uc = fd.units
        logger.debug("new limits : %s", str(uc[-1].ppm_limits()))
        logger.debug("new number of point %d", fd_.data.shape[-1])

        fd.forceupdatemodel = True
        
        return
 
#===============================================================================
# phase
#===============================================================================
def cursor2max(fd):
    """
    Find and set phcursor the position of the ideal reference peak for the 1st order phasing
    i.e., the maximum peak position. In order to avoir problems dur to sometimes 
    distorsions at the edge of the spectrum (digital filter problems), the search 
    is limited to the central portion of the spectrum
    
    Parameters
    ----------
    fd : Source instance
        Source instance 
        
    """
    if not fd.is_freq[-1]:
        # this search is not relevant for FID's
        return 0 
    
    s = list(fd.data.shape)
    si = s[-1]
    s[-1] = si/2 # we will work only on the half center
    lb = si/4.
    ub = 3.*lb
    data = fd.data[...,lb:ub]   #TODO: may be nicer to work with numpy masked array
    data = np.abs(data)
    cursor = np.argmax(data)  # this should give a flattened index on multidimensionnal array
    cursor = np.unravel_index(cursor, s)[-1] # so we unravel the indexes
    #set the fd attributes
    fd.phcursor = cursor + si/4. # do not forget that the indexwas found from the si/4 position

        
def pk(fd, p0=0, p1=0):
    """
    linear phase correction, work always on the last axis (axis=-1)
    If other axis need to be phased, first transpose the array
    The 1st order correction is calculated with respect to the 
    phcursor position (automatically set to the position of the 
    maximum peak intensity.
    
    Parameters
    ----------
    fd : Source instance
        Source instance 
    p0 : float, optional
        zero order phase correction, added to the value stored in fd
    p1 : float, optional
        1st order phase correction, added to the value stored in fd

    Examples
    --------
        
        >>> from masai.api import *
        >>> fd = Source(path = 'test/1')
        >>> ft(fd)
        >>> pk(fd)
        
        or for 2D
        
        >>> fd= Source(path = 'test/2')
        >>> ft(fd)
        >>> pk(fd)

   
    """ 
    if not fd.is_freq[-1]:
        # this apply only to fourier transformed data
        return  
    
    logger.debug("pk")
    if p0!=0 :  fd.PHC0[-1] = p0 + fd.PHC0[-1]
    if p1!=0 :  fd.PHC1[-1] = p1 + fd.PHC1[-1]
    
    p0 = fd.PHC0[-1] - fd.PHC0SAV[-1] 
    p1 = fd.PHC1[-1] - fd.PHC1SAV[-1] 
    if p0==0.0 and p1==0.0:
        return

    fd.data = _pk(fd.data, p0, p1)
    
    fd.PHC0SAV[-1] = fd.PHC0[-1]
    fd.PHC1SAV[-1] = fd.PHC1[-1]
        
    return

def apk(fd, delta=2.0):
    """ 
    Automatic linear phase correction (zero order phase only)

    .. todo:: 
    
        improve the choice of the processing area (parameter delta to change)
        
    Parameters
    ----------
    fd : Source instance
        Source instance 
    delta : float, optional, default=2
        Fraction of the maximum intensity to determine the optimization region 
            


    Examples
    --------
        
        >>> from masai.api import *
        >>> fd = Source(path = 'test/1')
        >>> ft(fd)
        >>> res = apk(fd)
        
        or for 2D
        
        >>> fd= Source(path='test/2')
        >>> ft(fd)
        >>> res2 = apk(fd)

    """
    if not fd.is_freq[-1]:
        # this apply only to FT transformed data
        return
    
    logger.debug("apk")
    # error function
    def phascalc(p, s, intervalle): 
        p0 = p[0]
        sc = s.copy()
        scp = _pk(sc, p0)[intervalle]
        return  np.arctan(np.sum(scp.imag) / np.sum(scp.real)) ** 2 # / np.sum(np.abs(sc.real) ** 2)

    pk(fd)
    
    n = fd.data.size / 4
    pwr = np.abs(fd.data)
    mi = np.unravel_index(np.argmax(pwr), fd.data.shape)[-1]
    s = fd.data[mi - n / 2:mi + n / 2]
    pwr = pwr[mi - n / 2:mi + n / 2]
    
    m = np.max(pwr)
    intervalle = pwr > (m / delta) 
        
    p0 = np.rad2deg(np.arctan(np.sum(s.imag) / np.sum(s.real)))
    p = (p0,)
    
    sys.stdout = Blackhole() #suppress output of the fmin function
    p0 = optimize.fmin(phascalc, p, args=(s, intervalle), xtol=1e-8)[0]
    sys.stdout = STDOUT # Restore normal output
    
    fd.PHC0[-1] = (fd.PHC0[-1]+p0) % 360.
    pk(fd)

    pmax = np.argmax(np.abs(fd.data.real))
    # pmax is found on a flattened array
    # so we use flatten() in the next line
    if fd.data.real.flatten()[pmax] < 0:   
        fd.PHC0[-1] = (fd.PHC0[-1]+180.) % 360.    
        pk(fd)
    
    logger.info("Automatic calculated phase was %.3f"%p0)
        
    return

def ft(fd):
    """
    Achieve the fourier transform (data should not be already transformed) 
    
    Note that an automatic zerofilling to the nearest power of two is automatically applied
    
    Parameters
    ----------
    fd : Source instance
        Contains bruker NMR parameters and  data

    Examples
    --------
        
        >>> from masai.api import *
        >>> fd = Source(path = 'test/1')
        >>> ft(fd)
        >>> ft(fd)
        False
        
        or for 2D
        
        >>> fd= Source(path='test/2')
        >>> ft(fd)

    """
    logger.debug("ft")
    
    if fd.is_freq[-1]:
        logger.warn("Data already transformed: use 'ift' instead?")
        return False
    
    # Zero filling
    si = fd.SI[-1] 
    zf(fd, si)
        
    # process ft
    fd.data = _ft(fd.data, si)
    fd.is_freq[-1] = True
    
    #we need to reset the position of the phcursor
    cursor2max(fd)
    
    return

def ift(fd):
    """
    Achieve the inverse fourier transform (data should be already transformed) 
    
    Note that an automatic zerofilling to the nearest power of two is automatically applied
    
    Parameters
    ----------
    fd : Source instance
        Contains bruker NMR parameters and  data
    


    Examples
    --------
        
        >>> from masai.api import *
        >>> fd = Source(path = 'test/1')
        >>> ft(fd)
        >>> ift(fd)
        
        or for 2D
        
        >>> fd= Source(path='test/2')
        >>> ft(fd)
        >>> ift(fd)

    """
    logger.debug("ift")
     
    if not fd.is_freq[-1]:
        logger.warn("Data not yet transformed: use 'ft' instead?")
        return False
    
    # process ft
    fd.data = _ift(fd.data)
    fd.data[fd.data.shape[-1]/2:] = 0.0
    fd.is_freq[-1] = False
    
    return

def efp(fd, autophase=False, delta=20):
    """
    Process sequential em, ft, pk bruker-like commands

    Parameters
    ----------
    fd : Source instance
        Contains bruker NMR parameters and  data
    autophase : bool, optional, default = False
        True if apk is applied instead of pk for automatic
        phasing
    delta : float, optional, default=20 
        (see apk function description for more information)  

    Examples
    --------
        
        >>> from masai.api import *
        >>> fd = Source(path = 'test/1')
        >>> efp(fd)
        
        or for 2D
        
        >>> fd= Source(path='test/2')
        >>> efp(fd)
        >>> efp(fd)
        False
        
    """
    logger.debug("efp")
    
    if fd.is_transposed:
        logger.warn("Data are transposed, we cannot use this procedure.")
        return False
    
    if fd.is_freq[-1]:
        logger.warn("The data are already transformed along F2")
        return False
    
    tdeff(fd) 
    em(fd) 
    ft(fd)
    
    if not autophase:
        pk(fd)
    else:
        apk(fd, delta)
    
    return
    
def xf2(fd, autophase=False, delta=20):
    """
    Process sequential em, ft, pk bruker-like commands (alias for efp, more specific for 2D spectra)

    
    Parameters
    ----------
    fd : Source instance
        Contains bruker NMR parameters and  data
    autophase : bool, optional, default = False
        True if apk is applied instead of pk for automatic
        phasing
    delta : float, optional, default=20 
        (see apk function description for more information)
        

    Examples
    --------
        
        >>> from masai.api import *
        >>> fd = Source(path = 'test/1')
        >>> xf2(fd)
        
        or for 2D
        
        >>> fd= Source(path='test/2')
        >>> xf2(fd)
   
    """     

    logger.debug("xf2:...")
    if fd.is_transposed: 
        logger.warn("the data are transposed, we cannot use this procedure (which apply only to indirect dimension")
        return False
    
    return efp(fd, autophase = autophase, delta = delta)

def xf1(fd, autophase=False, delta=20., filling=0,  shear=0.0, shift=0.0):
    """
    Fourier transform of the indirect dimension of a 2D NMR spectra

    Parameters
    ----------
    fd : Source instance
        Contains bruker NMR parameters and data
    autophase : bool, optional, default = False
        True if apk is applied instead of pk for automatic
        phasing
    delta : float, optional, default=20
        (see apk function description for more information)
    shear : Float, optional, default=0.0                
        If shear has a non-zero float value, it gives the shearing ratio to use
    shift: Float
        Amount of shifting in the F1 dimension in Hz
        


    Examples
    --------
        
        It is not possible for a 1D spectrum:
        
        >>> from masai.api import *
        >>> fd = Source(path = 'test/1')
        >>> xf2(fd)
        >>> xf1(fd)
        False
        
        For 2D:
        
        >>> fd= Source(path='test/2')
        >>> xf2(fd)
        >>> xf1(fd)
        >>> 
   
    """
    logger.debug("xf1:...")
    if fd.is_transposed:
        logger.warn('This procedure work only on non-transposed data.')
        return False
    
     
    if fd.PULPROG[-1].find('Multiplex') != -1 and fd.PARMODE[-1] == 1:
        # this is a specific case where the data where originally
        # acquired in 3D mode and processed with the MSM program
        # to give a 2D spectrum. In this case (this has to be changed in the MSM
        # processing, FnMODE and MC2 are wrong.
        # The mode of acquisition is in principle STATES, so we change it here
        fd.FnMODE[-1] = undefined
        fd.MC2[-1] = STATES
        
    mode = fd.FnMODE[-1]
    if mode == undefined:
        mode = fd.MC2[-1]
        
    if mode == STATES:  
        makestates(fd)
        
    else:
        # not yet implemented
        logger.warn("NotImplementedError")
        return False
    
    # phase for shearing and shift in F1
    if shear!=0.0 or shift!=0.0:
        logger.debug("shearing with ratio:%.5f"%shear)
        sw = fd.SW_h[0]  # remember that the table is transposed (direct dimension is now 0).
        si = fd.data.shape[0]
        in0 = fd.IN[0][0]
        p1 = -2.*3.14159*shear*sw*in0/si
        
        # additional F1 shift ?
        p0= -2.*3.14159*shift*in0;
        
        # create a 2D array of phases and performs the phase correction
        phases= np.fromfunction(lambda i0, i1: np.exp((p1*(i0-si/2)+p0)*i1*1.0j), fd.data.shape, dtype=float)
        fd.data = fd.data*phases
        
#    if ( mq<0 ) ph1= -ph1;
#    d1= fabs(ratio - (double)mq);
#    sfo11= d1*sfo1;

    if filling>0:
        zf(fd, filling)
         
    ft(fd)
    
    if not autophase:
        pk(fd)
    else:
        apk(fd, delta)
            
    #transpose back
    fd = transpose(fd)
    
    return

#===============================================================================
# utility functions
#===============================================================================
def findregion(xdata, region):
    """Return an array of indice corresponding to a given region
    
    xdata: ndarray
        array of abscisses (or ordinates) where to select a region
    region: list of float
        (low, high) list in any order giving the bound of the region 
    """
    _low = min(region)
    _high = max(region)
    region = np.argwhere((xdata <= _high) & (xdata >= _low)).ravel() 
    return list(region)
    
def largest_power_of_2(value):
    """
    Find the largest and nearest power of two of a value
    
    Parameters
    ----------
    value: int
    
    Returns
    -------
    pw: int
        Power of 2. 
        
    Examples
    --------
    
        >>> largest_power_of_2(1020)
        1024
        >>> largest_power_of_2(1025)
        2048
        >>> largest_power_of_2(0)
        2
        >>> largest_power_of_2(1)
        2
        
        
    """
    value = max(2, value)
    p = int(pow(2, np.ceil(np.log(value) / np.log(2))))
    return p

def transpose(fd):
    """Transpose 
    
    Parameters
    ----------
    fd : Source instance
        Contains bruker NMR parameters and data
    
    """
    if fd.ndims <2: 
        return False
    fd.is_transposed = not fd.is_transposed
    fd.data = np.transpose(fd.data)
    return

def makestates(fd):
    """Transforms according to the States procedure. 
    The data are transposed.
    
    Parameters
    ----------
    fd : Source instance
        Contains bruker NMR parameters and data
    
    """

    fd.is_transposed = True
    ar = np.transpose(fd.data[::2]) # transposed real part 
    ai = np.transpose(fd.data[1::2]) # transposed imag part
    fd.data = ar.real + ai.real * 1.0j
    return

def shearing_ratio(fd, p=3, q=0):
    """
    Calculation of the MQMAS or STMAS experiments shearing ratio
    
    For MQMAS, the shearing ratio is 
    
    .. math:: 
        k =\\frac{1}{9} \\frac{p \\left( -17{{p}^{2}}+36 I(I+1)-10 \\right)}{4 I( I+1)-3}
    
    For STMAS, the shearing ratio is
    
    .. math::
        k=\\frac{17}{3}\\frac{q^2}{\\left( 4I\\left( I+1 \\right)-3 \\right)}-1
    
    Examples
    --------
    
        >>> from masai.api import Source, shearing_ratio
        >>> f = Source(path='test/1')
        >>> shearing_ratio(f)
        1.5833333333333333
        >>> shearing_ratio(f, 5)
        2.0833333333333335
        
        
    Parameters
    ----------
    fd : Source instance
        Contains the required parameters
    p : Int, optional 
        default = 3 (3Q)
        multi-quantum order of the selected coherence (if mq==1 and q!=0 then a STMAS experiment is assumed, 
        if mq==2 then a STMAS Double_Quantum experiment is assumed instead. In all other case (where odd mq>2) 
        then MQMAS experiment is assumed. 
    q : Int, optional
        satelitte order for STMAS.
                
    returns
    -------
    k : float
        Ratio corresponding to the MQMAS shearing
        
    """

    I = fd.NUC1[-1].spin
    nspin = int(np.ceil(2.0 * I))
    p = float(abs(p))    
    q = float(q)
    
    if p == nspin:
        p = -p
    
    if p > nspin:
        logger.warn('coherence order p cannot be greater than 2*I')
        return 0
    
    if p > 2 or p == -nspin: 
        # MQMAS experiment assumed
        ratio = (p / 9.0) * (-17.0 * (p ** 2) + 36.0 * I * (I + 1.) - 10.0) / (4.0 * I * (I + 1.0) - 3.0)
        
    elif p == 1 and q != 0:
        # STMAS experiment assumed
        ratio = -((17.0 / 3.0) * (q ** 2) / (4.0 * I * (I + 1.0) - 3.0) - 1.0)
        
    return ratio
    


#! /usr/bin/env python
#===============================================================================
# modelling.parameters
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================

"""
Model parameters handling

"""
#===============================================================================
# python import
#===============================================================================
import os
import sys
import re                       # For regular expression search
import string
import types
import logging
import importlib
import numpy as np
from UserDict import UserDict   # This is to be able to create a special dictionary

#===============================================================================
# enthought import
#===============================================================================
from traits.api import (HasTraits, Str, Instance, on_trait_change,
                        Event, Property, cached_property, Button)
from traitsui.api import (View, Item, CodeEditor, HGroup)

#===============================================================================
# local import
#===============================================================================
from models import list_of_models, list_of_baselines
from masai.utils import DEnum
from masai.utils import Timer

#===============================================================================
# logging
#===============================================================================
logger = logging.getLogger()

#===============================================================================
# constants
#===============================================================================
FIXED=True

doc_params = """
#-----------------------------------------------------------
# syntax for parameters definition:
# name: value, low_bound,  high_bound
#  * for fixed parameters
#  $ for variable parameters
#  > for reference to a parameter in the COMMON block
#    (> is forbidden in the COMMON block)
# common block parameters should not have a _ in their names
#-----------------------------------------------------------
"""

default_params_script = """
COMMON:
# common parameters ex.
# $ gwidth: 1.0, 0.0, none
# $ gratio: 0.5, 0.0, 1.0

"""
#===========================================================================
# id_generator
#===========================================================================
def _id_generator():
    """Returns a sequence of numbers for the title of the objects.
    
    Example:
    --------
    
    >>> id_generator.next()
    1
    
    """
    n=1
    while True:
        yield n
        n += 1
id_generator = _id_generator()

#===============================================================================
# FitParameters
#===============================================================================
#TODO: Traitify
class FitParameters(UserDict):
    """
    Allow passing a dictionary of parameters with additional properties
    to the fit function. Check if the parameter is between the specified bounds
    if any.
    
    Examples
    --------
    
    >>> fp = FitParameters()
    
    """
        
    def __init__(self):
        UserDict.__init__(self)     # Create a dictionary class
        self.lob = {}               # Lower bound
        self.upb = {}               # Upper bound
        self.fixed = {}             # true for non-variable parameter
        self.reference = {}         # 
        self.model = {}             # model to use
        self.models = []            # list of models
        self.sequence = ''          # sequence used in the experiment
        
    def __setitem__(self, key, value):
        key = str(key)
        if not self.reference.has_key(key):
            self.reference[key] = False
        if self.reference[key]:
            # we get a reference to another parameter 
            self.data[key] = str(value)
            self.fixed[key] = True 
        elif isinstance(value,tuple) or isinstance(value,list):
            self.data[key] = self._evaluate(value[0])
            self.lob[key] = None
            self.upb[key] = None
            try:
                if len(value)>2:
                    self.lob[key] = self._evaluate(value[1])
                    self.upb[key] = self._evaluate(value[2])
                    self._checkerror(key)
            except:
                pass
            self.fixed[key] = False
            if isinstance(value[-1],bool):
                self.fixed[key] = value[-1]
        else:
            self.data[key] = self._evaluate(value)
            self.lob[key] = None
            self.upb[key] = None
            self.fixed[key] = False

    def __getitem__(self, key):
        key = str(key)
        if self.data.has_key(key):
            return self.data[key]
        raise KeyError("parameter %s is not found"%key)

    def iteritems(self):
        return self.data.iteritems()

    def _checkerror(self, key):
        key = str(key)
        if self.lob[key] is None and self.upb[key] is None:
            return False
        elif (self.lob[key] is not None and self.data[key] < self.lob[key]) \
            or (self.upb[key] is not None and self.data[key] > self.upb[key]) :
            raise ValueError('%s value %s is out of bounds'%(key, str(self.data[key])))
        
    def __str__(self): 
        message = "#PARAMETER SCRIPT\n\nCOMMON: \n"
        for key in self.keys():
            if "_" not in  key: 
                # this is probably a common parameter 
                # (outside a model definition)
                if self.fixed[key]:
                    keystring = "\t* %s" % key
                else:
                    keystring = "\t$ %s" % key 
                lob = self.lob[key]
                upb = self.upb[key]
                if lob <= -0.1 / sys.float_info.epsilon:
                    lob = "none" 
                if upb >= +0.1 / sys.float_info.epsilon:
                    upb = "none"
                val = str(self.data[key])
                message += "%s: %s, %s, %s \n" % (keystring, val, lob, upb)
        message += ""             
        models = self.models  
        for model in models:
            message += "\nMODEL: %s\n" % model
            message += "shape: %s\n" % self.model[model]
            for key in self.keys():
                ends = "_%s" % (model)
                if not key.endswith(ends): 
                    continue 
                keystring = key[:-len(ends)] 
                
                if self.reference[key]:
                    message += "\t> %s: %s" % (keystring, self.data[key])
                    
                else:
                    if self.fixed[key]:
                        keystring = "\t* %s" % keystring
                    else:
                        keystring = "\t$ %s" % keystring    
                    lob = self.lob[key]
                    upb = self.upb[key]
                    if lob <= -0.1 / sys.float_info.epsilon:
                        lob = "none" 
                    if upb >= +0.1 / sys.float_info.epsilon:
                        upb = "none"
                    val = str(self.data[key])
                    message += "%s: %s, %s, %s" % (keystring, val, lob,  upb)
                
                message += "\n"          
        return message

    def _evaluate(self, strg):
        """
        Allow the evaluation of strings containing some operations
        
        Parameters
        ----------
        strg: string  
            A string to evaluate containing multiplier, 
            e.g., '10 k' evaluate to 10 000.
        
        Return
        ------
        value : float or bool
            Value of the string, or False, if there is an error
        
        """
        res = False
        
        if type(strg) is types.StringType or type(strg) is types.UnicodeType:
            #strg=string.upper(strg)
            p = re.compile('\s+')
            m = p.split(string.strip(strg))
            try:
                res = eval(m[0])
            except NameError:
                message="Cannot evaluate '"+strg+"' >> "+m[0]+" is not defined"
                raise NameError,message
            except SyntaxError:
                message="Syntax error in '"+strg+"'"
                raise SyntaxError,message
            #read mulitplier
            if len(m) > 1:
                try:
                    res=res*eval(m[1])
                except NameError:
                    message= "Cannot evaluate '"+strg+"' >> "+m[1]+" is not defined"
                    raise ValueError,message
                except SyntaxError:
                    message="Syntax error in '"+strg+"'"
                    raise ValueError,message
        else:
            # not a string (probably a scalar that can be return as it is)
            res = strg
            
        return res
    
#===============================================================================
class FitSteps(UserDict):
#===============================================================================
    """
    Define which parameters are used in a step of the fit
    """
    count = 0
    def __init__(self, fitpar):
        UserDict.__init__(self)
        self.fixed = {}
        self.parameters = fitpar.keys

    def __setitem__(self, step, value):        
        self.data[step]=self.parameters
        self.fixed[step]={}

        # we start with all parameters fixed
        for k in self.parameters:
            self.fixed[step][k] = True

        if not value:
            value='allfixed'

        if isinstance(value,tuple) or isinstance(value,list):
            value = "|".join(value)

        if isinstance(value,str):
            if value[0]!="^": '^('+value
            if value[-1]!="$": value+")$"

            p=re.compile(value)
            for k in self.parameters:
                m = p.search(k)
                if m: self.fixed[step][k] = False
        else:
            raise ValueError, "FitStep["+str(step)+"]: Can't reconize this input!"

    def __getitem__(self, step):
        return self.fixed[step]

#===============================================================================
# ParameterScript
#===============================================================================
class ParameterScript(HasTraits):
    """
    This class allow some manipulation of the parameter list for modelling
    """
    fd = Instance('masai.fileio.source.Source')
    fp = Instance(FitParameters)
        
    scriptfilename = Str('')
    content = Str('')
    savecontent = Str('')
    
    _list_of_models = Property
    modeltype = DEnum(values_name='_list_of_models')
    addmodel = Button('ADD LINE')

    _list_of_baselines = Property
    basetype = DEnum(values_name='_list_of_baselines')
    addbase = Button('ADD BASELINE')  
      
    updated = Event   
    
    lastparfailure = Instance(Exception,allow_none=True) 
     
    timer = Instance(Timer)
    
    #===========================================================================
    # properties
    #===========================================================================
    @cached_property
    def _get__list_of_models(self):
        """
        Return a list of available model type, eg. voigtmodel ...
        """
        return list_of_models

    @cached_property
    def _get__list_of_baselines(self):
        """
        Return a list of available model type, eg. voigtmodel ...
        """
        return list_of_baselines
    
    #===========================================================================
    # view
    #===========================================================================
    def default_traits_view(self):        
        return View(
#                Group( 
                      Item('content',
                                  show_label=False,
                                  style='custom',
                                  #height=400,
                                  editor=CodeEditor(auto_set=True),
                        ),
                        HGroup(
                              Item('addmodel', 
                                   show_label=False,),
                              Item('modeltype', 
                                   label='model type',),
                              visible_when='modeltype is not None and not lastparfailure'
                        ),
                        HGroup(
                              Item('addbase', 
                                   show_label=False,),
                              Item('basetype', 
                                   label='baseline',),
                              visible_when='basetype is not None and not lastparfailure'
                        ),
                        Item('lastparfailure',
                             show_label=False, 
                             visible_when='lastparfailure', 
                             style='readonly', 
                             emphasized=True),
#                    ),
                    resizable=True,
                )
        
    @on_trait_change('fd')
    def _fdchanged(self):
        """
        This event is fired when the source is first defined or changed
        """
        logger.debug("initialisation of parameter class")
        
        #try to load a previously open parameter script file
        self.scriptfilename = os.path.join(self.fd._directory, "param.fit")
        if os.path.exists(self.scriptfilename):
            #if it is found, set content  
            with open(self.scriptfilename, 'rb') as f:
                self.content = f.read()
        else:
            #else, we load some default script
            self.content = default_params_script
        
        # define a timer
        if not self.timer:
            self.timer = Timer(3.0, self._check_parameters)
            self.timer.start()
            
    def _check_parameters(self):
        """
        Check the validity of the parameters
        """
        try:
            self.lastparfailure = None
            if self.content != self.savecontent:
                logger.debug("parameters: content updated")
                self.fp = self._interpret(self.content)
                self._save_parameters()
                self.updated = True
            self.savecontent = self.content
            return True
        except Exception,e:
            self.lastparfailure = e
            logger.error(e)
            return False
        
    def _save_parameters(self):
        """
        Save the content to a file
        """   
        if self.scriptfilename and not self.lastparfailure:  
            logger.debug("parameters: content saved")
            f = open(self.scriptfilename, 'wb')
            f.write(self.content)
            f.close()            

#    @on_trait_change('content')
#    def _update_parameters(self, name, old, new):
#        """
#        fired if some changes are detected in the script content
#        """        
#        if self._check_parameters():
#            logger.debug("parameters: content updated")
#            self._save_parameters()   
         
    def _addmodel_fired(self, baseline=False):
        """
        Add action fired. We import the required model and some parameters
        before writing in the script content
        """
        self.fd.show_model=True
        
        module = importlib.import_module("masai.modelling.models")
        if not baseline:
            class_ = getattr(module, self.modeltype)
        else:
            class_ = getattr(module, self.basetype)
        script = class_().script
        
        # we determine some parameters from the source data
        uc = self.fd.units
        x = uc[-1].ppm_scale() # scale in ppm
        y = self.fd.data.real # np.abs(self.fd.data[lx/4:-lx/4])
        try:
            ym = self.fd.modeldata[-1]
            y = y-ym
        except:
            pass
        ampl = np.max(y)
        argm = np.argmax(y) #+lx/4
        pos = x[argm]
        poslb, poshb = uc[-1].ppm_limits()
        w= np.argwhere(y > ampl/2.)
        wmin = np.min(w) #+lx/4
        wmax = np.max(w) #+lx/4
        xm = min( abs(x[wmax] - pos), abs(x[wmin] - pos))
        width = xm * 2. # sigma corresponding to the FWHM
#        #correct the ampl to take into account the width
        delta = abs(x[1]-x[0])
        ampl = ampl * np.pi * width /2. /delta
        newid = id_generator.next()
        #try to get the amplitude for polynomial baseline
        a = y[0]
        b = y[np.argmin(y)]
        scale = abs(a-b)/abs(x[0]-x[np.argmin(y)])**2/ampl
        name = 'line'
        if baseline:
            name='baseline'
        while '%s%d'%(name,newid) in self.fp.models:
            newid = id_generator.next()
        self.content += "\n" + script % {"id":newid, "width":width, "ampl":ampl, 
                         "pos":pos, "poslb":poslb, "poshb":poshb, 'scale':scale}

    def _addbase_fired(self):
        """ Add baseline action fired. We import the required model and some parameters
        before writing in the script content
        """
        self._addmodel_fired(baseline=True)
        
    def _interpret(self, script):
        """
        Interpreter of the script content 
        """
        # init some flags
        modlabel = None
        common = False
        fixed = False
        reference = False
        
        # create a new FitParameters instance
        fp = FitParameters()
        
        # start interpreting
        lines = script.split('\n')
        lc = 0
        for item in lines:
            lc += 1  # count the lines
            line = item.strip()
            if line == '' or line.startswith("#"):
                # this is a blank or comment line, go to next line
                continue
            #split around the semi-column
            s = line.split(':')
            if len(s) != 2:
                raise SyntaxError('Cannot interpret line %d : A semi-column is missing?' % lc)
                 
            key, values = s
            key = key.strip().lower()
            if key.startswith('model'):
                modlabel = values.lower().strip()
                if modlabel not in fp.models:
                    fp.models.append(modlabel) 
                common = False
                continue
            elif key.startswith('common') or key.startswith('vars'): 
                common = True
                modlabel = 'common'
                continue
            elif key.startswith('shape'):
                shape = values.lower().strip()
                if shape is None or (shape not in list_of_models and shape not in list_of_baselines):
                    raise SyntaxError('Shape of this model "%s" was not specified or is not implemented'%shape)
                fp.model[modlabel] = shape
                common = False
                continue
            else:
                if modlabel is None and not common:
                    raise SyntaxError('The first definition should be a label for a model or a block of variables or constants.')                      
                # get the parameters
                if key.startswith('*'):
                    fixed = True
                    reference = False
                    key = key[1:].strip() 
                elif key.startswith('$'):
                    fixed = False
                    reference = False
                    key = key[1:].strip()
                elif key.startswith('>'):
                    fixed = True
                    reference = True
                    key = key[1:].strip()
                else:
                    raise SyntaxError('Cannot interpret line %d: A parameter definition must start with *,$ or >' % lc)
                    
                # store this parameter
                s = values.split(',')
                if len(s) > 3:
                    raise SyntaxError('line %d: value, min, max shoud be defined in this order' % lc)
                    s = s[:3]
                elif len(s) == 2:
                    raise SyntaxError('only two items in line %d' % lc)
                    s.append('none')
                elif len(s) == 1:
                    s.extend(['none', 'none'])
                value, mini, maxi = s
                if mini.strip().lower() in ['none', '']: mini = str(-1. / sys.float_info.epsilon)
                if maxi.strip().lower() in ['none', '']: maxi = str(+1. / sys.float_info.epsilon)
                if modlabel != 'common':
                    ks = "%s_%s" % (key, modlabel)
                else:
                    ks = "%s" % (key)  
                fp.reference[ks] = reference
                if not reference:
                    fp[ks] = value.strip(), mini.strip(), maxi.strip(), fixed      
                else:
                    fp[ks] = value.strip()
        return fp 

                



#! /usr/bin/env python
#===============================================================================
# version
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================
"""
This module contains version information. It use git to get the last version and
revision number.

"""
import commands

def get_git_version():
    """Get the current version  of the local git repository using
    the 'git describe' shell command.
    """
    status,output = commands.getstatusoutput('git describe')
    if status==0:
        return output.split('-')
    else:
        return '0.1','','' 

_TITLE_= u'Masai, a framework for processing and modelling of solid state NMR spectra'
_COPYRIGHT_ = u"2012-2013, C.Fernandez @ LCS (ENSICAEN/University of Caen/CNRS)"

_VERSION_, _REVISION_, _GITID_ = get_git_version()



#$ proc_1D
import pylab as pl

# You must always import the needed function or classes
from masai.api import Source, efp, pk, apk
# or do a wild import using:
#from masai import Source, efp
# Uncomment the above line to use it.

# read the data and open a new instance of Source
spc = Source('test/1')
fig1 = pl.figure(1,figsize=(10,4))      # to create a figure

# make some processings
spc.LB[-1] = 50.         # set line broadening
efp(spc)                 # make em, ft and pk at once
# you can add an automatic phase correction of 
# the newly (not well phased) data
apk(spc)
# or phase it manually
spc.PHC0[-1] = 1.        # set the 0-order phase parameter in degrees
spc.PHC1[-1] = 0.        # set the 1-order phase parameter in degrees
pk(spc)                  # apply the phase

# now plot it
data = spc.data.real     # take only the real part
pl.plot(data, 'k')       # create a plot in fig1 with a black trace
pl.show()                # show the fig1              




#! /usr/bin/env python
#===============================================================================
# masai.utils.tools
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================
""".. _tools:

This :mod:`tools` module contains useful methods or classes used 
in other modules...

"""

import threading
import time

class Timer:
    def __init__(self, tempo, target, args= [], kwargs={}):
        self._target = target
        self._args = args
        self._kwargs = kwargs
        self._tempo = tempo

    def _run(self):
        self._timer = threading.Timer(self._tempo, self._run)
        self._timer.start()
        self._target(*self._args, **self._kwargs)
        
    def start(self):
        self._timer = threading.Timer(self._tempo, self._run)
        self._timer.start()

    def stop(self):
        self._timer.cancel()

if __name__ == '__main__':

    pass


#$ proc_1D

from masai.api import Source
import pylab as pl

# read the data and open a new instance of Source

spc = Source('test/1')
fig1 = pl.figure(1, figsize=(10,3))      # to create a figure
data = spc.data.real     # take only the real part
pl.plot(data, 'k')       # create a plot in fig1 with a black trace
data = spc.data.imag     # take now the imaginary part
pl.plot(data, 'r')       # add it to the plot with a red trace
pl.xlim(0, 300)          # shows only the first 300 points
pl.show()                # show the fig1              



"""
 py2app/py2exe build script for Masai.

 Usage (Mac OS X):
     python setup.py py2app

 Usage (Windows):
     python setup.py py2exe
"""

import sys
from setuptools import find_packages
from numpy.distutils.core import setup  

from masai.version import _TITLE_, _VERSION_, _REVISION_ 

with open('README.rst') as fp:
    LONG_DESCRIPTION = fp.read()

with open('doc/source/changelog.rst') as fp:
    LONG_DESCRIPTION += '\n\n' + fp.read()

DISTNAME = 'masai'
VERSION = "%s.%s" % (_VERSION_, _REVISION_)
DESCRIPTION = _TITLE_
AUTHOR = 'Christian Fernandez'
EMAIL = 'christian.fernandez@ensicaen.fr'
URL = 'http://fernandezc.github.com/masai'
DOWNLOADURL = 'http://pypi.python.org/pypi/masai/'
LICENSE = 'CeCILL-B'

CLASSIFIERS = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: Other/Proprietary License',
        'Programming Language :: Python',
        'Programming Language :: Fortran',
        'Topic :: Scientific/Engineering',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux'
        ]    

mainscript = 'masai/masaigui.py'

if sys.platform == 'darwin':
    extra_options = dict(
        setup_requires=['py2app', 'numpy'],
        app=[mainscript],
        # Cross-platform applications generally expect sys.argv to
        # be used for opening files.
        options=dict(py2app=dict(argv_emulation=True)),
        
    )
elif sys.platform == 'win32':
    extra_options = dict(
        setup_requires=['py2exe', 'numpy'],
        app=[mainscript],
    )
else:
    extra_options = dict(
        # Normally unix-like platforms will use "setup.py install"
        # and install the main script as such
        setup_requires=['numpy'],
        scripts=[mainscript],
        
    )

def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    pulsar_dir = 'masai/modelling/pulsar'
    config = Configuration(DISTNAME,
                           parent_package,
                           top_path,
                           namespace_packages=['masai'],
                           version=VERSION,
                           author=AUTHOR,
                           author_email=EMAIL,
                           description=DESCRIPTION,
                           license=LICENSE,
                           url=URL,
                           long_description=LONG_DESCRIPTION)
    config.add_data_dir(('modelling/script','masai/modelling/script'))
    config.add_data_dir(('images','masai/images'))
    config.add_extension('forpulsar',
                  sources=['%s/f90/forpulsar.f90' % pulsar_dir,
                  '%s/f90/forpulsar.pyf' % pulsar_dir],
                  f2py_options=['--quiet'])
    return config

setup(configuration=configuration,
      download_url=DOWNLOADURL,
      classifiers=CLASSIFIERS,
      keywords='NMR, Python, Fortran',
      packages=find_packages(),
      include_package_data=True,
      install_requires=['numpy'],
      zip_safe=True,
      requires=['numpy', 'scipy', 'matplotlib', 'traits'],
      entry_points={
        'console_scripts': [
            "masai = masai.masaigui:run",
        ],
      },
      **extra_options
      )



#! /usr/bin/env python
#===============================================================================
# modelling.pulsar.pypulsar
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================

""" docstring """

from masai.forpulsar import (parameters, reset, compute)
                                    #reset_parameters, reset_operators, reset_share, reset, 
                                    #diagonalize, compute)

import logging
logger = logging.getLogger()

#def WRITE_LOG(level=None,strg=None):
#    """
#    Print a message from the FORPULSAR module
#    Callback function called by Fortran
#    
#    Parameters
#    ----------
#    level: integer 
#         Level of the debugging message
#         "1":"critical",
#         "2":"error",
#         "3":"warning",
#         "4":"info",
#         "5":"log",
#         "6":"debug"
#    strg: string 
#        message to display
#   
#    """
#    strtype={
#         "1":"critical",
#         "2":"error",
#         "3":"warning",
#         "4":"info",
#         "5":"debug",   # because log does not work
#         "6":"debug"
#    }
#
#    try:
#        LogMethod = getattr(logger, "%s" % strtype[str(level)])
#        LogMethod('%s: %s'%(strtype[str(level)],str(strg)))
#    except:
#        pass
    
if __name__ == '__main__':
    # some tests
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.WARNING)

    import numpy as np

#    for I in np.arange(1.5, 4.56, 1.):
#        reset()
#        #diagonalize.dealloc()
#        parameters.protonfrequency = 400000000.
#        parameters.idealpulse= True #False #True
#        parameters.pulse=[[1, .000001, 20000, 0, 0, 0, 0, 0]]
#        parameters.delay=[[1, 0.0, 1]]
#        parameters.chemicalshift = [[1, 0, 0, 0, 0, 0, 0]]
#        parameters.quadrupole = [[1, 4e6, 0., 0, 0, 0]]
#        parameters.t2  = [[1, 1e6]]
#        parameters.spinningangle = 54.736
#        parameters.spinningspeed = 00000.
#        parameters.qfactor = 0.000001
#        parameters.rfstep = 4
#        parameters.npts = 8192
#        parameters.nsb = 0
#        parameters.accuracy = 16
#        parameters.sw = 5000000. #parameters.spinningspeed
#        parameters.sr = 0
#        parameters.nall = 1
#        parameters.rcph = 0.
#
#        parameters.nucleus = [[1,                #index
#                           I,                #spin
#                           100000000.0,      #larmor
#                           1.0]]           #abundance
#    
#        import time
#        start = time.time()
#        done = compute(WRITE_LOG)
#        end = time.time()
#        print end-start
#        
#        import pylab as pl
#        pl.plot(parameters.spec.real)
#        
#        s=np.zeros((2*I,))
#        for i,m in enumerate(np.arange(-I, I-.1, 1.)):
#            s[i] = I*(I+1.)-m*(m+1.)
#        
#        print I, s, s.sum(), parameters.spec.real.sum()
#        print [(2*I+1)/2,(2*I+1)/2-1], parameters.ros[(2*I+1)/2,(2*I+1)/2-1]
#        
#    pl.show()
    
    reset()
    
    parameters.protonfrequency = 400000000.
    parameters.idealpulse= True #False #True
    parameters.pulse=[[1, .0, 0., 0, 0, 0, 0, 0]]
    parameters.delay=[[1, 0.0, 1]]
    parameters.t2  = [[1, 1e6]]
    parameters.spinningangle = 54.736
    sw = 50000.
    parameters.spinningspeed = sw
    parameters.qfactor = 0.000001
    parameters.rfstep = 4
    npts = 8192
    parameters.npts = npts
    parameters.nsb = 0
    parameters.accuracy = 16
    parameters.sw = sw
    parameters.sr = 0
    parameters.nall = 1
    parameters.rcph = 0.

    import pylab as pl
    import time

    CA = lambda I : (I*(I+1.)-3./4.)/(2.*I*(2.*I-1))**2
    A = lambda I,cq, w0 : CA(I)*cq**2/w0
    cq2w0fromA = lambda I,A : A/((I*(I+1.)-3./4.)/(2.*I*(2.*I-1))**2)
    
    import os
    pathdb = os.path.join(os.path.expanduser('~'), '.masai','db')

    I=1.5
    w0=100.0e6
    iter = np.arange(5000,35000,10000)
    specall = np.zeros((iter.size, npts))
    start = time.time()
    for i, Ax in enumerate(iter):
        cq = (cq2w0fromA(I,Ax)*w0)**.5
        print cq
        qis = -(I*(I+1.)-3./4.)*(3.*cq/(2.*I*(2.*I-1)))**2/w0/30
        parameters.quadrupole = [[1, cq, 0., 0, 0, 0]]
        parameters.chemicalshift = [[1, -qis, 0, 0, 0, 0, 0]]
        parameters.nucleus = [[1,      #index
                               I,      #spin
                               w0,     #larmor
                               1.0]]   #abundance
        done = compute()
        specall[i]=parameters.spec.real
    np.save('res', specall)
    end = time.time()-start
    print 'time', end
  
    start = time.time()
    done = compute()
    end = time.time()-start
    print 'time direct', end
    
    start = time.time()
    specall = np.load('res.npy')
    spec = specall[-1]
    end = time.time()-start
    print 'time file', end
    
    #pl.plot(specall.T)
    #pl.show()
    
    x = np.arange(npts)-npts/2
    pl.plot(x,specall.T)
#    pl.show()
    
    x2= (iter[1]/iter[0])*x
    s = specall[0]/2.
    pl.plot(x2,s,':')
    pl.show()
#    pl.contour(specall)
#
#    #2D interpolation
#    a,b = specall.shape 
#    print a,b   
#    from scipy import ogrid, sin, mgrid, ndimage, array
#
#    x,y = np.ogrid[0:a:1,0:b:1]
#    print x, y
#    
#    newx,newy = np.mgrid[0:a:0.1,0:b:1]
#    print "newx", newx
#    print "newy", newy
#    
#    x0 = x[0,0]
#    y0 = y[0,0]
#    dx = x[1,0] - x0
#    dy = y[0,1] - y0
#    ivals = (newx - x0)/dx
#    jvals = (newy - y0)/dy
#    coords = array([ivals, jvals])
#    
#    print coords
#    
#    newf = ndimage.map_coordinates(specall, coords)
#    pl.contour(newf)
#    pl.show()
#     
   


#! /usr/bin/env python
# -*- coding: utf-8 -*-
#===============================================================================
# modelling.models
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================
"""This module holds the defintions all the various models 
used by Masaï.

"""
#===============================================================================
# logger
#===============================================================================
import logging
logger = logging.getLogger()
import os
import numpy as np
from masai.processing.processbase import _ft
from masai.modelling.pulsar.pypulsar import (parameters, compute, reset)

list_of_baselines = [ # baseline models
                   'polynomialbaseline',
                ]
seq_ind_models = [ #list of sequence independant models
                'voigtmodel',
                'gaussianmodel',
                'lorentzianmodel',
                ]
list_of_models = seq_ind_models
list_of_models.extend(
                      [
                      'quad2model',
                      'quad2ssbmodel',
                      'quadallssbmodel',
                      #'czjzekquad2model'
                      ]
                      )

pathdb = os.path.join(os.path.expanduser('~'), '.masai','db')

#===============================================================================
# PolynomialBaseline
#===============================================================================
class polynomialbaseline(object):
    """
    Arbitrary-degree polynomial (degree limited to 10, however).
    As a linear baseline is automatically calculated, this polynom is always of
    greater or equal to order 2 (parabolic function). 
    
    .. math::
        f(x) = ampl * \\sum_{i=2}^{max} c_i*x^i
        
    """    
    args = ['ampl']
    args.extend(['c_%d'%i for i in range(2,11)])
    
    script = """MODEL: baseline%(id)d\nshape: polynomialbaseline
    # This polynom starts at the order 2 
    # as a linear baseline is additionnaly fitted automatically
    # parameters must be in the form c_i where i is an integer as shown below
    * ampl: %(scale).3g, 0.0, None
    * c_2: 1.0, None, None
    * c_3: 0.0, None, None
    * c_4: 0.0, None, None
    # etc...
    """ 
    def f(self,x, ampl, *c_, **kargs): 
        c=[0.0,0.0]
        c.extend(c_)
        return ampl*np.polyval(np.array(tuple(c))[::-1],x-x[x.size/2])
    
#===============================================================================
# #===============================================================================
# # Gaussian2DModel
# #===============================================================================
# class gaussian2dmodel(object):
#    """
#    Two dimensional Gaussian model (*not* normalized - peak value is 1).
#    
#    .. math::
#        A e^{\\frac{-(x-\\iso_x)^2}{2 \\gb_x^2}} e^{\\frac{-(y-\\iso_y)^2}{2 \\gb_y^2}}
#    
#    """
#    args = ['amp','gbx','gby','posx','posy']
#    def f(self, xy, amp, gbx, gby, posx, posy, **kargs):
#        gbx = float(gbx)
#        gby = float(gby)
#        x,y = xy
#        xo = x-posx
#        xdenom = 2*gbx*gbx
#        yo = y-posy
#        ydenom = 2*gby*gby
#        return amp*np.exp(-xo*xo/xdenom-yo*yo/ydenom)
#===============================================================================

#===============================================================================
# GaussianModel
#===============================================================================
class gaussianmodel(object):
    """
    Normalized 1D gaussian function:
    
    .. math::
        f(x) = \\frac{ampl}{\\sqrt{2 \\pi \\sigma^2} } \\exp({\\frac{-(x-pos)^2}{2 \\sigma^2}})
    
    where :math:`\\sigma = \\frac{width}{2.3548}`
    
    """
    args = ['ampl', 'width', 'pos']
    script = """MODEL: line%(id)d\nshape: gaussianmodel
    * ampl: %(ampl).3f, 0.0, None
    * width: %(width).3f, 0.0, None
    * pos: %(pos).3f, %(poslb).3f, %(poshb).3f
    """ 
    def f(self,x,ampl,width,pos, **kargs):
        gb = width/2.3548
        tsq=(x-pos)*2**-0.5/gb
        w = np.exp(-tsq*tsq)*(2*np.pi)**-0.5/gb
        w = w * (x[1]-x[0])
        return ampl*w
    
#===============================================================================
# LorentzianModel
#===============================================================================
class lorentzianmodel(object):
    """
    A standard Lorentzian function (also known as the Cauchy distribution):
    
    .. math::
        f(x) = \\frac{ampl * \\lambda}{\\pi [(x-pos)^2+ \\lambda^2]}
    
    where :math:`\\lambda = \\frac{width}{2}` 
    
    """
    args = ['ampl', 'width', 'pos']
    script = """MODEL: line%(id)d\nshape: lorentzianmodel
    * ampl: %(ampl).3f, 0.0, None
    * width: %(width).3f, 0.0, None
    * pos: %(pos).3f, %(poslb).3f, %(poshb).3f
    """ 
    def f(self, x, ampl, width, pos, **kargs):
        lb = width/2.
        w = lb/np.pi/(x*x-2*x*pos+pos*pos+lb*lb)
        w = w * (x[1]-x[0])
        return ampl*w
        
#===============================================================================
# VoigtModel
#===============================================================================
class voigtmodel(object):
    """
    A Voigt model constructed as the convolution of a :class:`GaussianModel` and
    a :class:`LorentzianModel` -- commonly used for spectral line fitting.
    
    """
    args = ['ampl', 'width', 'ratio','pos']
    script = """MODEL: line%(id)d\nshape: voigtmodel
    * ampl: %(ampl).3f, 0.0, None
    * width: %(width).3f, 0.0, None
    * pos: %(pos).3f, %(poslb).3f, %(poshb).3f
    * ratio: 0.0, 0.0, None
    """ 
    def f(self, x, ampl, width, ratio, pos, **kargs):
        from scipy.special import wofz
        gb = ratio * width /2.3548
        lb = (1.-ratio) * width /2.
        if gb < 1.e-16:
            return lorentzianmodel().f(x, ampl, lb*2., pos, **kargs)
        else:
            w=wofz(((x-pos)+1.0j*lb)*2**-0.5/gb)
            w = w.real*(2.*np.pi)**-0.5/gb
            w = w * (x[1]-x[0])
            return ampl*w
        
#===============================================================================
# Quad2model
#===============================================================================
class quad2model(object):
    """Second order quadrupolar lineshape assuming infinite MAS spinning speed.
    (work also for second-order quadrupolar lineshape of static powder.
     
    The lineshape is calculated using Pulsar. 
    
    """
    args = ['ampl', 'width', 'ratio', 'pos', 'cq', 'etaq']
    script = """MODEL: line%(id)d\nshape: quad2model
    * ampl: %(ampl).3f, 0.0, None
    * width: %(width).3f, 0.0, None
    * pos: %(pos).3f, %(poslb).3f, %(poshb).3f
    * ratio: 0.0, 0.0, None
    * cq: 1.0, 0.0, None
    * etaq: 0.0, 0.0, 1.0 
    """ 
    def f(self, x, ampl, width, ratio, pos, cq, etaq, 
                    vr=None,  czjzek=False, all=False, **kargs):
        # evaluate all parameters
        reset()      
        if kargs.has_key('sequence'):
            sequence = kargs['sequence']
            if sequence.type !='ideal_pulse':
                pass #logger.error('sorry non-ideal pulse not yet implemented')
        fd = kargs['fd']
        nucleus = fd.NUC1[-1]
        
        sw_h = float(fd.SW_h[-1])
        nsb=0
        if vr is None:  # this is different from 0 (static) 
            vr= sw_h
            nsb = 0
        elif vr>0.0:
            nsb = int(sw_h/vr/2.)
            #here we must limit the number of ssb in order to avoid a crackh of pulsar
            # for small spinning speed
            nsb = min(30, nsb)   #TODO: let the user choose the best max number of spinning sidebands
        fratio = 5.5856912e6/nucleus.gamma
        sfo1 = fd.SFO1[-1]
        sf = fd.SF[-1]
        sr = (sfo1 - sf) * 1.e6
        protonf = fratio * sfo1
        si = fd.data.shape[-1]
                
        # set parameters for the pulsar calculation
        logging.disable(logging.INFO)  # to avoid multiple output during calculation
        parameters.protonfrequency = protonf
        parameters.nucleus = [[1,                #index
                           nucleus.spin,         #spin
                           sfo1*1.e6,            #larmor
                           nucleus.abundance]]   #abundance
        parameters.idealpulse= True
        parameters.pulse=[[1, 2, 0, 0, 0, 0, 0, 0]]
        parameters.delay=[[1, 0.0, 1]]
        pos = pos * sfo1 - sr
        parameters.chemicalshift = [[1, pos, 0, 0, 0, 0, 0]]
        parameters.t2  = [[1, 1e6]]
        parameters.spinningangle = 54.74
        parameters.spinningspeed = vr
        parameters.qfactor = 0.001
        parameters.rfstep = 5
        parameters.npts = si
        parameters.nsb = nsb
        parameters.accuracy = 8
        parameters.sw = sw_h
        parameters.sr = 0
        parameters.nall = 1
        if all:
            parameters.nall=0
        parameters.rcph = 0. #TODO: change this in the f90 compute function
                
        if not czjzek:
            parameters.quadrupole = [[1, cq, etaq, 0, 0, 0]] 
            compute()
            data = parameters.spec.real
        else:
            parameters.accuracy = 3    
            sigmaq = cq
            #look for the max of the distribution
            vq = np.linspace(0.0,6.*sigmaq,128)
            mp = np.max(pczjzeck(vq, .6, sigmaq))
            #now we will use a gauss legendre quadrature (may be not the best?)
        
        broad = voigtmodel().f(x, 1.,width,ratio, sr/sfo1)   # centered 
                                                # on the middle of the window!
        data = np.convolve(data,broad,'same')
        logging.disable(logging.NOTSET)  # to restore normal logger level
        
        return ampl*data    # voir pour la normalisation


#===============================================================================
# quad2modelssb
#===============================================================================
class quad2ssbmodel(object):
    """Second order quadrupolar lineshape assuming finite MAS spinning speed.
    (spinning sidebands are calculated in this model)
     
    The lineshape is calculated using Pulsar. 
    
    """

    args = ['ampl', 'width', 'ratio', 'pos', 'cq', 'etaq','vr']
    script = """MODEL: line%(id)d\nshape: quad2ssbmodel
    * ampl: %(ampl).3f, 0.0, None
    * width: %(width).3f, 0.0, None
    * pos: %(pos).3f, %(poslb).3f, %(poshb).3f
    * ratio: 0.0, 0.0, None
    * cq: 1.0, 0.0, None
    * etaq: 0.0, 0.0, 1.0 
    * vr: 12000., 0.0, None
    """ 
    def f(self, x, ampl, width, ratio, pos, cq, etaq, vr, **kargs):
        return quad2model().f(x, ampl, width, ratio, pos, cq, etaq, vr, **kargs)

#===============================================================================
# quadallmodelssb
#===============================================================================
class quadallssbmodel(object):
    """All transition lineshape calculated at the first 
    and second order quadrupolar 
    assuming finite MAS spinning speed. 
    It therefore include the spinning sidebands.
     
    It is calculated using Pulsar. 
    
    """

    args = ['ampl', 'width', 'ratio', 'pos', 'cq', 'etaq','vr']
    script = """MODEL: line%(id)d\nshape: quadallssbmodel
    * ampl: %(ampl).3f, 0.0, None
    * width: %(width).3f, 0.0, None
    * pos: %(pos).3f, %(poslb).3f, %(poshb).3f
    * ratio: 0.0, 0.0, None
    * cq: 1.0, 0.0, None
    * etaq: 0.0, 0.0, 1.0 
    * vr: 12000., 0.0, None
    """ 
    def f(self, x, ampl, width, ratio, pos, cq, etaq, vr, **kargs):
        return quad2model().f(x, ampl, width, ratio, pos, cq, etaq, 
                                                          vr, all=True, **kargs)
        
#===============================================================================
# #===============================================================================
# # czjzekquad2model
# #===============================================================================
# class czjzekquad2model(quad2model):
#    args = ['ampl', 'width', 'ratio', 'pos', 'sigmaq']
#    script = """MODEL: line%(id)d\nshape: czjzekquad2model
#    * ampl: %(ampl).3f, 0.0, None
#    * width: %(width).3f, 0.0, None
#    * pos: %(pos).3f, %(poslb).3f, %(poshb).3f
#    * ratio: 0.0, 0.0, None
#    * sigmaq: 0.0, 0.0, None
#    """ 
#    def f(self, x, ampl, width, ratio, pos, sigmaq, **kargs):
#        return quad2model.f(self, x, ampl, width, ratio, pos, sigmaq, 
#                                                      0.0, czjzek=True, **kargs)
#===============================================================================

def pczjzeck(vq, eta, sigma):
    """
    Czjzek probability distribution function [#]_:

    .. math::
       P(\\nu_Q, \\eta, \\sigma) = \\frac{1}{\\sqrt{2 \\pi} \\sigma^5 } \\nu_Q^4 \\eta 
       \\left(1-\\frac{\\eta^2}{9} \\right) \\exp \\left[
       -\\frac{\\nu_Q^2 \\left( 1 + \\frac{\\eta^2}{3} \\right)}{2\\sigma^2}\\right]

       
    where
   
    .. math::
        \\nu_Q = \\frac{3 C_Q}{2 I (2 I - 1)}  
         
    .. [#] Jean-Baptiste d'Espinose de Lacaillerie, Christian Fretigny, Dominique Massiot, 
        MAS NMR spectra of quadrupolar nuclei in disordered solids: The Czjzek model, 
        Journal of Magnetic Resonance, Volume 192, Issue 2, June 2008, 
        Pages 244-251, ISSN 1090-7807, 10.1016/j.jmr.2008.03.001.
        (`<http://www.sciencedirect.com/science/article/pii/S1090780708000888>`_)
   
    """
    p = 1. /( np.sqrt(2.* np.pi) * sigma**5)
    p = p * vq**4 * eta *(1. - eta**2 / 9.)
    p = p * np.exp(- vq**2 * (1. + eta**2 /3. )/2./ sigma**2 )
    return p    

#===============================================================================
# simpsonquad2model
#===============================================================================
class simpsonquad2model(object):
    """Second order quadrupolar lineshape assuming MAS spinning speed 
    equal to the spectral width.
    
    It is calculated using Simpson. (EXPERIMENTAL... need more developpement
    for a good integration in Masai)
    
    It requires that Simpson is installed separately and accessible from a shell
    in the the masai main directory [#]_.
    
    .. [#] Simpson can be downloaded here : `Simpson <http://bionmr.chem.au.dk/bionmr/software/simpson.php>`_
    
    """

    args = ['ampl', 'width', 'ratio', 'pos', 'cq', 'etaq']
    script = """MODEL: line%(id)d\nshape: simpsonquad2model
    * ampl: %(ampl).3f, 0.0, None
    * width: %(width).3f, 0.0, None
    * pos: %(pos).3f, %(poslb).3f, %(poshb).3f
    * ratio: 0.0, 0.0, None
    * cq: 1.0, 0.0, None
    * etaq: 0.0, 0.0, 1.0 
    """ 
    def f(self, x, ampl, width, ratio, pos, cq, etaq, **kargs):
        gb = width * ratio 
        lb = width * (1.-ratio)        
        if kargs.has_key('sequence'):
            sequence = kargs['sequence']
            if sequence.type !='ideal_pulse':
                pass #logger.error('sorry non-ideal pulse not yet implemented')
        fd = kargs['fd']
        nucleus = fd.NUC1[-1]
        sw_h = fd.SW_h[-1]
        vr = sw_h
        fratio = 5.5856912e6/nucleus.gamma
        sfo1 = fd.SFO1[-1]
        sf = fd.SF[-1]
        sr = (sfo1 - sf) * 1.e6/sfo1 # en ppm
        protonf = fratio * sfo1
        si = fd.data.shape[-1]
        lb = lb * sfo1
        gb = gb * sfo1
        r = gb/(gb+lb)
        
        path = os.path.join(os.path.expanduser('~'),'.masai')
        temp_in = os.path.join(path, 'temp.in')
        temp_fid = os.path.join(path, 'temp.fid')

        # in this script the vr is made exactly the spectral width
        simpson_script = \
        """
        spinsys {
              channels %(symbol)s
              nuclei   %(symbol)s
              shift 1 %(pos).3fp 0 0 0 0 0
              quadrupole 1 2 %(cq).3f %(etaq).3f 0 0 0
        }

        par {  
              proton_frequency %(protonf).0f
              method          gcompute  
              gamma_angles    1  
              crystal_file    rep2000
              start_operator  I1x
              detect_operator I1c
              spin_rate       %(vr).3f
              np              %(td)d
              sw              spin_rate
              variable tsw    1e6/sw
              verbose         0
        }

        proc pulseq {} {
          delay 9999
        }

        proc main {} {
          global par
          set f [fsimpson]
          faddlb $f %(lb)s %(r)s 
          fsave $f %(path)s -raw_bin
          
        }
        
        """%{'symbol':nucleus, 'cq':cq, 'vr':vr, 'etaq':etaq, 'protonf': protonf, 
             'td':2048, 'si':si, 'lb':lb+gb, 'r':r, 'pos':pos-sr, 'path':temp_fid}

        with open(temp_in, 'wb') as f:
            f.write(simpson_script)
        import subprocess
        subprocess.check_output( ['simpson',temp_in])
        
        data = read_raw_bin(temp_fid)
        try:
            intensity = np.absolute(data[0])  
        except IndexError:
            raise IndexError("""Cannot find the first point of the fid.
            Simpson probably failed to execute the script""")
        np.save('temp', data/intensity)
        data = _ft(data/intensity, si).real     
        return ampl*data    # voir pour la normalisation
        

#===============================================================================
# getmodel
#===============================================================================
def getmodel(x, modelname, par, **kargs):
    """Get the model for a given x vector.
    
    Parameters
    -----------
    x : ndarray
        Array of frequency where to evaluate the model values returned by the
        f function.
    modelname : str
        name of the model class to use.
    par : :class:`Parameters` instance
        parameter to pass to the f function
    kargs: any
        Keywords arguments to pass the the f function
        
    Returns
    -------
    ndarray : float
        an array containing the calculated model.
        
    """
    model = par.model[modelname]
    modelcls = globals()[model]
    
    # take an instance of the model
    a = modelcls()
    
    # get the parameters for the given model
    args = []
    for p in a.args:
        try:
            args.append(par['%s_%s'%(p, modelname)])
        except KeyError as e:
            if p.startswith('c_'):
                # probably the end of the list due to a limited polynomial degree
                pass
            else:
                raise KeyError(e.message)
            
    return a.f(x,*args, **kargs)

def read_raw_bin(filename):
    """
    Read a SIMPSON file wich has been saved using:: 
    
        fsave $f filename -raw_bin
    
    ..note::
        adapted from Jonathan J. Helmus (jjhelmus@gmail.com)
        Version: 0.1 (2012-04-13)
        License: GPL
    
    Returns
    --------
        ndarray : complex
            a 1D complex array

    """
    data = np.fromfile(filename, dtype='float32')
    half = int(data.shape[0] / 2)
    cdata = np.empty( (half, ), dtype='complex64')
    cdata.real = data[:half]
    cdata.imag = data[half:]
    return cdata


if __name__ == '__main__':
    pass